from .image_annotation import AnnotatedImage, ImageAnnotater, Drawing

import tempfile
import os
import time

import huggingface_hub
from ultralytics import YOLO
import numpy as np
import cv2
import boto3


class ImageAutoAnnotater(ImageAnnotater):
    def __init__(self, image, resize_constant, model):
        super().__init__(image, resize_constant)
        self.model = model

    def annotate(self):
        prepared_image = self._get_prepared_image(self.image)
        xmin, xmax, ymin, ymax = self._get_roi(prepared_image)

        cropped_image, mask = self.auto_annotate(prepared_image, xmin, xmax, ymin, ymax)
        return AnnotatedImage(cropped_image, mask)
        

    def auto_annotate(self, prepared_image, xmin, xmax, ymin, ymax):
        window = prepared_image[ymin:ymax, xmin:xmax, :]

        results = self.model(window)

        window = cv2.cvtColor(window, cv2.COLOR_BGR2GRAY)

        mask = []

        if not results[0].masks:
            return window, np.zeros(window.shape, np.uint8)

        for r in results:
            plot = r.plot(boxes=False)
            for x, y in r.masks.xy[0]:
                mask.append(int(x))
                mask.append(int(y))

        ctr = np.array(mask).reshape((-1, 1, 2)).astype(np.int32)

        mask = np.zeros(window.shape, np.uint8)
        cv2.drawContours(mask, [ctr], -1, (255), 1)

        mask = Drawing(window, mask).draw()        

        return window, mask

    def _get_prepared_image(self, image):
        shape = image.shape
        image = cv2.resize(image, (shape[1] * self.resize_constant, shape[0] * self.resize_constant), cv2.INTER_NEAREST)
        return image


class AutoAnnotater:
    def __init__(self, resize_constant, endpoint_name, access_key, secret_key):
        self.resize_constant = resize_constant
        self.model = Model(endpoint_name, access_key, secret_key)

    def annotate(self, image):
        return ImageAutoAnnotater(image, self.resize_constant, self.model).annotate()

class Model:
    def __init__(self, endpoint_name, access_key, secret_key):
        self.endpoint_name = endpoint_name
        self.sagemaker_client = boto3.client("sagemaker-runtime",
                                             aws_access_key_id=access_key,
                                             aws_secret_access_key=secret_key,
                                             region_name="us-east-1")

    def __call__(self, image: AnnotatedImage):
        with tempfile.TemporaryDirectory() as tmpdirname:
            file = os.path.join(tmpdirname, "image.jpg")
            cv2.imwrite(file, image)
            with open(file, "rb") as f:
                payload = f.read()
            out = self.sagemaker_client.invoke_endpoint(EndpointName=self.endpoint_name,
                                                        ContentType="application/x-image",
                                                        Body=payload)
            print(out)
            return out



import numpy as np
import matplotlib.pyplot as plt
import cv2

class UpdateablePlot:
    def __init__(self, num_subplots=1, height=3.2, typ='plot', title='Plots'):
        # valid types: 'scatter', 'plot'
        self.num_subplots = num_subplots
        self.height = height
        self.width = num_subplots * height * 1.2
        self.fig = plt.figure(figsize=(self.width, self.height))
        self.fig.tight_layout(rect=[0, 0, .5, 1])
        self.fig.suptitle(title)
        self.axes = []
        self.typ = typ
        self.num_subplots = num_subplots
        self.title = title
        for i in range(num_subplots):
            ax = self.fig.add_subplot(1, num_subplots, i+1)

            if typ == 'plot':
                data, = ax.plot([], [])
            elif typ == 'scatter':
                data = ax.scatter([], [])
            else:
                raise Exception("Invalid type")

            self.axes.append((ax, data))

        for i in range(num_subplots):
            self.change_subplot_dims(i)

        plt.close()

    def reset(self):
        self.fig = plt.figure(figsize=(self.width, self.height))
        self.fig.tight_layout()
        self.fig.suptitle(self.title)
        self.axes = []
        for i in range(self.num_subplots):
            ax = self.fig.add_subplot(1, self.num_subplots, i+1)

            if self.typ == 'plot':
                data, = ax.plot([], [])
            elif self.typ == 'scatter':
                data = ax.scatter([], [])
            else:
                raise Exception("Invalid type")

            self.axes.append((ax, data))

        for i in range(self.num_subplots):
            self.change_subplot_dims(i)

        plt.close()


    def update_subplot(self, subplot_idx, xval, yval, padding_ratio):
        if subplot_idx not in range(len(self.axes)):
            raise Exception("Invalid index")

        ax, data, = self.axes[subplot_idx]

        if self.typ == 'plot':
            xdata, ydata = data.get_data()
            xdata = list(xdata)
            ydata = list(ydata)
            xdata.append(xval)
            ydata.append(yval)
            data.set_data(xdata, ydata)

            max_val = max(ydata) if ydata else 0
            min_val = min(ydata) if ydata else -0
            xlim = (xdata[0], xdata[-1])

        elif self.typ == 'scatter':
            points = data.get_offsets().data
            new_data = np.array([[xval, yval]])
            points = np.append(points, new_data, axis=0)
            data.set_offsets(points)
            max_val = max(points, key=lambda b : b[1])[1]
            min_val = min(points, key=lambda b : b[1])[1]
            xlim = (points[0][0], points[-1][0])
        else:
            raise Exception("Invalid type")
        
        datarange = max_val - min_val
        max_val = max_val + (padding_ratio / 2 * datarange)
        min_val = min_val - (padding_ratio / 2 * datarange)
        ylim = (min_val, max_val)

        if ylim[0] == ylim[1]:
            ylim = (1, -1)

        if xlim[0] == xlim[1]:
            xlim = (1, -1)

        ax.set_ylim(ylim)
        ax.set_xlim(xlim)


    def set_subplot_chars(self, subplot_idx, title, xlabel, ylabel, fontsize=10):
        if subplot_idx not in range(len(self.axes)):
            raise Exception("Invalid index")

        ax, data, = self.axes[subplot_idx]

        ax.set_title(title, fontsize=fontsize)
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)

    def change_subplot_dims(self, subplot_idx, xlim=(0,10), ylim=(0,10)):
        if subplot_idx not in range(len(self.axes)):
            raise Exception("Invalid index")

        ax, data, = self.axes[subplot_idx]
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

    def get_img(self):
        fig = self.fig
        fig.canvas.draw()
        img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        img  = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = cv2.cvtColor(img,cv2.COLOR_RGB2BGR)
        return img



import cv2
import numpy as np

import os

class Util:

    def get_dip_values(contour, frame, scaling_factor, um_per_pixel=1):
        binary_image = cv2.cvtColor(np.zeros_like(frame), cv2.COLOR_BGR2GRAY)

        cv2.drawContours(binary_image, [contour], -1, (1), thickness=cv2.FILLED)

        binary_image_dip = dip.Image(binary_image.astype(np.uint8))
        binary_image_dip.Convert('BIN')

        labeled_image = dip.Label(binary_image_dip)
        measurement = dip.MeasurementTool.Measure(labeled_image, features=['Perimeter', 'SolidArea', 'Roundness'])
        perimeter = measurement['Perimeter'][1][0]
        area = measurement['SolidArea'][1][0]
        roundness = measurement['Roundness'][1][0]

        print(perimeter, measurement['Perimeter'], area)

        p, a = perimeter / scaling_factor * um_per_pixel, area / pow(scaling_factor, 2) * pow(um_per_pixel, 2)

        #return p, a, math.pi * 4 * a / pow(p, 2)
        return p, a, roundness



    def get_perimeter(contour, scaling_factor, um_per_pixel=1):
        raw = cv2.arcLength(contour, True)
        return raw / scaling_factor * um_per_pixel

    def get_area(contour, scaling_factor, um_per_pixel=1):
        raw = cv2.contourArea(contour)
        return raw / pow(scaling_factor, 2) * pow(um_per_pixel, 2)

    def get_circularity(contour, frame, scaling_factor, um_per_pixel=1):
        # area = Util.get_area(contour, scaling_factor, um_per_pixel)
        # perimeter = Util.get_perimeter(contour, scaling_factor, um_per_pixel)
        # return math.pi * 4 * area / pow(perimeter, 2)
        binary_image = cv2.cvtColor(np.zeros_like(frame), cv2.COLOR_BGR2GRAY)

        cv2.drawContours(binary_image, [contour], -1, (1), thickness=cv2.FILLED)

        binary_image_dip = dip.Image(binary_image.astype(np.uint8))
        binary_image_dip.Convert('BIN')

        labeled_image = dip.Label(binary_image_dip)

        measurement = dip.MeasurementTool.Measure(labeled_image, features=['Roundness'])

        roundness = measurement['Roundness'][1][0]
        return roundness


    def get_z_height(contour, scaling_factor, um_per_pixel=1):
        x, y, w, h = cv2.boundingRect(contour)
        return h * um_per_pixel / scaling_factor

    def combine_images(img1, img2, pad_color=(0,0,0)):
        w1 = img1.shape[1]
        w2 = img2.shape[1]

        if w1 > w2:
            return cv2.vconcat([img1, Util.pad_image(w1, img2)])
        else:
            return cv2.vconcat([Util.pad_image(w2, img1), img2])

    def pad_image(new_width, frame, pad_color=(0, 0, 0)):
        width, height = frame.shape[1], frame.shape[0]

        if width > new_width:
            raise Exception("New width too small")

        width_to_add = new_width - width
        wta_perside = width_to_add // 2

        blank_image = np.zeros((height,wta_perside,3), np.uint8)

        if width_to_add % 2 == 0:
            img = cv2.hconcat([blank_image, frame, blank_image])
        else:
            blank_image_2 = np.zeros((height, wta_perside + 1, 3), np.uint8)
            img = cv2.hconcat([blank_image, frame, blank_image_2])

        return img

    def plot_to_opencv(fig):
        fig.canvas.draw()
        img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        img  = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = cv2.cvtColor(img,cv2.COLOR_RGB2BGR)
        return img

    def get_window(frame, centerX, window_width):
        new_left = max(0, centerX - window_width)
        new_right = min(frame.shape[1], centerX + window_width)

        if new_left == 0:
            new_right = min(frame.shape[1], window_width)
        elif new_right == frame.shape[1]:
            new_left == max(0, frame.shape[1] - window_width)

        new_img = frame[:, new_left:new_right]

        return new_img

    def list_pt(path):
        return [ f for f in os.listdir(path) if f.endswith(".pt") ]



from collections import deque
import os

from .weight_manager import WeightManager
from .data_handler import DataHandler
from .util import Util
from .image_processing import ProcessedImage

import cv2
import tqdm
import numpy as np
import pandas as pd

class InferencePipeline:
    
    def __init__(self, model, framerate, window_width, scaling_factor, um_per_pixel, output_folder):
        self.model = model
        self.framerate = framerate
        self.scaling_factor = scaling_factor
        self.um_per_pixel = um_per_pixel
        self.output_folder = output_folder
        self.window_width = window_width

        self.process_queue = deque()


    def process_video(self, video_path, scatter=False, verbose=False):
        video_name = os.path.basename(video_path)
        output_video_name = os.path.join(self.output_folder, os.path.splitext(video_name)[0] + "-analysis" + ".avi")
        output_csv_name = os.path.join(self.output_folder, os.path.splitext(video_name)[0] + "-analysis" + ".csv")

        dh = DataHandler(deltatime=1/self.framerate, scatter=scatter)

        cap = cv2.VideoCapture(video_path)

        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)) * self.scaling_factor
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)) * self.scaling_factor

        centerX = width

        frame = np.zeros((height, width, 3), np.uint8)
        out = Util.combine_images(frame, dh.plot.get_img())
        num_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        video = cv2.VideoWriter(output_video_name, cv2.VideoWriter_fourcc(*'MJPG'), 15, (out.shape[1], out.shape[0]))
        
        for _ in tqdm.tqdm(range(num_frames)):
            ret, frame = cap.read()

            if not ret:
                break

            frame = cv2.resize(frame, (width, height), cv2.INTER_NEAREST)

            img = ProcessedImage(frame, centerX, self.window_width, self.scaling_factor, self.um_per_pixel, self.model)
            dh.update_data(area=img.get_parameter('area'),
                           perimeter=img.get_parameter('perimeter'),
                           height=img.get_parameter('height'),
                           circularity=img.get_parameter('circularity'),
                           centerX=img.get_parameter('centerX'))

            if img.get_contour() is not None:
                cv2.drawContours(frame, [img.get_contour()], -1, (0, 255, 0), 1)

            if dh.prev_data['centerX']:
                centerX = max(dh.prev_data['centerX'], self.window_width)

            out = Util.combine_images(frame, dh.plot.get_img())
            video.write(out)

            pd.DataFrame(dh.data).to_csv(output_csv_name)
            
        cap.release()
        cv2.destroyAllWindows()
        video.release()

        return dh.data



import cv2

class Drawing:
    def __init__(self, image, mask):
        self.image = image
        self.mask = mask.copy()
        
        self.state = {
            "drawing": False,
            "white": False,
            "show_mask": False,
            "show_contours": True
        }

        self.window_name = "Draw Contour"

    def draw(self):
        cv2.namedWindow(self.window_name, cv2.WINDOW_KEEPRATIO)
        cv2.resizeWindow(self.window_name, (1920, 1080))
        cv2.setMouseCallback(self.window_name, self._draw_circle)
        
        while True:
            self._draw_frame()
            key = cv2.waitKey(1)

            if key == 27:
                break
            elif key == ord('m'):
                self.state["show_mask"] = not self.state["show_mask"]
            elif key == ord('c'):
                self.state["white"] = not self.state["white"]
            elif key == ord('o'):
                self.state["show_contours"] = not self.state["show_contours"]


        cv2.destroyWindow(self.window_name)

        return self.mask

    def _draw_frame(self):
        if self.state["show_mask"]:
            cv2.imshow(self.window_name, self.mask)
            return

        frame = self.image.copy()
        if self.state["show_contours"]:
            contours, hierarchy = cv2.findContours(self.mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
            cv2.drawContours(frame, contours, -1, (0, 255, 0), 1) 

        cv2.imshow(self.window_name, frame)


    def _draw_circle(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            self.state["drawing"] = True
        elif event == cv2.EVENT_LBUTTONUP:
            self.state["drawing"] = False
            
        if self.state["drawing"]:
            color = (255, 255, 255) if self.state["white"] else (0, 0, 0)
            cv2.circle(self.mask, (x, y), 3, color, -1)



from ..plotting import UpdateablePlot

class DataHandler:

    def __init__(self, deltatime=1, scatter=False):

        self.time = 0
        self.deltatime = deltatime
        self.scatter = scatter

        self.data = {
            "area": [],
            "height": [],
            "perimeter": [],
            "velocity": [],
            "acceleration": [],
            "circularity": [],
            "time": []}

        self.prev_data = {
            'area' : 0,
            'perimeter' : 0,
            'height' : 0,
            'velocity' : 0,
            'acceleration' : 0,
            'circularity' : 0,
            'centerX' : None}

        self.plot = UpdateablePlot(6, 3.2, 'scatter') if scatter else UpdateablePlot(6, 3.2, 'plot')
        self.plot.set_subplot_chars(0, "Area", "Time (us)", "Area (um^2)")
        self.plot.set_subplot_chars(1, "Perimeter", "Time (us)",  "Perimeter (um)")
        self.plot.set_subplot_chars(2, "Height", "Time (us)", "Height (um)")
        self.plot.set_subplot_chars(3, "Velocity", "Time (us)", "Velocity (um/s)")
        self.plot.set_subplot_chars(4, "Acceleration", "Time (us)", "Acceleration (um/s^2)")
        self.plot.set_subplot_chars(5, "Circularity", "Time (us)", "Circularity")


    def update_data(self, area, perimeter, height, circularity, centerX=None):

        self.time += self.deltatime
        self.data['time'].append(self.time)

        if centerX and self.prev_data['centerX']:
            velocity = (centerX - self.prev_data['centerX']) / self.deltatime
        else:
            velocity = 0

        acceleration = (velocity - self.prev_data['velocity']) / self.deltatime

        self.data['area'].append(area)
        self.prev_data['area'] = area
        self.data['height'].append(height)
        self.prev_data['height'] = height
        self.data['perimeter'].append(perimeter)
        self.prev_data['perimeter'] = perimeter
        self.data['velocity'].append(velocity)
        self.prev_data['velocity'] = velocity
        self.data['acceleration'].append(acceleration)
        self.prev_data['acceleration'] = acceleration
        self.data['circularity'].append(circularity)
        self.prev_data['circularity'] = circularity
        self.prev_data['centerX'] = centerX

        self.__update_plot()

    
    def get_plot_img(self):
        return self.plot.get_img()


    def __update_plot(self):
        self.plot.update_subplot(0, self.time, self.data['area'][-1], .1)
        self.plot.update_subplot(1, self.time, self.data['perimeter'][-1], .1)
        self.plot.update_subplot(2, self.time, self.data['height'][-1], .1)
        self.plot.update_subplot(3, self.time, self.data['velocity'][-1], .1)
        self.plot.update_subplot(4, self.time, self.data['acceleration'][-1], .1)
        self.plot.update_subplot(5, self.time, self.data['circularity'][-1], .1)



import os
import pathlib
import logging

import huggingface_hub

class Configs:
    
    temp_path = os.path.join(pathlib.Path(__file__).parents[1], "temp")
    weights_path = os.path.join(pathlib.Path(__file__).parents[1], "weights")

    if not os.path.isdir(temp_path):
        os.makedir(temp_path)

    if not os.path.isdir(weights_path):
        os.makedir(weights_path)

    huggingface_repo_id = "gt-sulchek-lab/cell-tracking"

    def huggingface_login(token, write_permission=False):
        try:
            huggingface_hub.login(token=token, write_permission=write_permission)

        except Exception as e:
            logging.error(e)
            return False

        return True

    def huggingface_logout():
        try:
            huggingface_hub.logout()

        except Exception as e:
            logging.error(e)
            return False

        return True



import os
import logging
import shutil

from .config import Configs
from .util import Util

import huggingface_hub
from ultralytics import YOLO

class WeightManager:

    available_models = Util.list_pt(Configs.weights_path)
    model = None

    def get_model():
        if not WeightManager.model:
            raise Exception("No model found")

        return WeightManager.model


    def select_current_model(model_name):
        if model_name in WeightManager.available_models:
            WeightManager.model = YOLO(os.path.join(Configs.weights_path, model_name))
        else:
            raise Exception("Model not found")
        

    def download_model_weights(huggingface_repo_id, model_name):
        try:
            huggingface_hub.hf_hub_download(repo_id=Configs.huggingface_repo_id,
                                            local_dir=Configs.weights_path,
                                            filename=model_name)

            if os.path.exists(os.path.join(Configs.weights_path, ".huggingface")):
                shutil.rmtree(os.path.join(Configs.weights_path, ".huggingface"))

        except Exception as e:
            logging.error(e)
            return False
        
        WeightManager.available_models = Util.list_pt(Configs.weights_path)
        return True


    def list_weights():
        return WeightManager.available_models


    def delete_weights(model_name):
        if model_name in WeightManager.available_models:
            os.remove(os.path.join(Configs.weights_path, model_name))
            WeightManager.available_models = Util.list_pt(Configs.weights_path)

        else:
            raise Exception("Model not found")


    def check_weights(model_name):
        return model_name in WeightManager.available_models



import time


class Configs:

    # IMAGE ANNOTATION CONFIGS

    current_time = time.localtime()
    
    # info
    year = current_time.tm_year
    version = "1.0"
    description = "not found"
    contributor = "not found"
    url = "not found"
    date_created = f"{current_time.tm_mday}-{current_time.tm_mon}-{current_time.tm_year}"

    default_info = {
        "year": year,
        "version": version,
        "description": description,
        "contributor": contributor,
        "url": url,
        "date_created": date_created
    }

    # license
    id = "not found"
    name = "not found"

    default_license = {
        "id": id,
        "name": name,
        "url": url
    }

    # annotation
    image_id = id
    category_id = "not found"
    segmentation = None
    area = None
    bbox = None
    iscrowd = 0

    default_annotation = {
        "id": id,
        "image_id": image_id,
        "category_id": category_id,
        "segmentation": segmentation,
        "area": area,
        "bbox": bbox,
        "iscrowd": iscrowd
    }
    
    # image
    width = 0
    height = 0
    file_name = "not found"

    default_image = {
        "id": id,
        "width": width,
        "height": height,
        "file_name": file_name
    }

    default_json = {
        "info": default_info,
        "images": [default_image],
        "annotations": [default_annotation],
        "licenses": [default_license]
    }

    def get_default_annotation():
        return Configs.default_annotation.copy()

    def get_default_image():
        return Configs.default_image.copy()

    def get_default_license():
        return Configs.default_license.copy()

    def get_default_info():
        return Configs.default_info.copy()

    def get_default_json():
        return Configs.default_json.copy()

    # OTHER CONFIGS

    default_roboflow_batch = "not found"



import json
import tempfile
import os

from .config import Configs
from .drawing import Drawing

import cv2
import roboflow
import numpy as np


class AnnotatedImage:
    def __init__(self, img, mask, filename="defaultfilename.png"):
        self.img = img
        self.mask = mask
        self.tempdir = tempfile.TemporaryDirectory()
        self.filename = os.path.join(self.tempdir.name, filename)
        self.json_filename = self.filename + "-annotation.json"

        self.annotation_json = Configs.get_default_json()
        self.annotation_json["images"] = [self._get_image(img)]
        self.annotation_json["annotations"] = [self._get_annotation(mask)]

        cv2.imwrite(self.filename, img)
        with open(self.json_filename, "w") as file:
            file.write(json.dumps(self.annotation_json))
        
    def __del__(self):
        self.tempdir.cleanup()

    def _get_annotation(self, mask):
        annotation = Configs.get_default_annotation()

        contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        segmentation = []
        for contour in contours:
            contour = contour.flatten().tolist()
            contour_pairs = [(contour[i], contour[i+1]) for i in range(0, len(contour), 2)]
            segmentation.append([int(coord) for pair in contour_pairs for coord in pair])

        area = cv2.contourArea(contours[0]) if contours else 0
        bbox = [int(x) for x in cv2.boundingRect(contours[0])] if contours else None

        annotation["segmentation"] = segmentation
        annotation["area"] = area
        annotation["bbox"] = bbox

        return annotation

    def _get_image(self, img):
        image = Configs.get_default_image()
        shape = img.shape

        image["width"] = shape[0]
        image["height"] = shape[1]
        image["file_name"] = self.filename

        return image

    def roboflow_upload(self, workspace, project, api_key=None):
        if api_key is None:
            api_key = input("Please input your roboflow api key here: ")
        rf = roboflow.Roboflow(api_key=api_key)
        project = rf.workspace(workspace).project(project)

        project.upload(image_path=self.filename,
                       annotation_path=self.json_filename,
                       batch_name=Configs.default_roboflow_batch)

    def show(self):
         contours, hierarchy = cv2.findContours(self.mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
         img = self.img.copy()
         cv2.drawContours(img, contours, -1, (0, 0, 0), 3)
         cv2.imshow("Annotation", img)
         while True:
            key = cv2.waitKey(1000)
            if key == 27 or cv2.getWindowProperty("Annotation", cv2.WND_PROP_VISIBLE) < 1:
                break
         cv2.destroyWindow("Annotation")

    def get_resized_image(self, window_width, offset):

        if self.annotation_json["annotations"]:
            seg = self.annotation_json["annotations"][0]["segmentation"][0]
            ctr = AnnotatedImage.coco_to_opencv(seg)
            M = cv2.moments(ctr)

            cX = int(M["m10"] / M["m00"])
            cY = int(M["m01"] / M["m00"])
        else:
            cX = random.randint(40, self.img.shape[1] - 40)
            cY = self.img.shape[0] // 2

        newLeft = max(0, offset + cX - window_width // 2)

        imwidth = self.img.shape[1]
        newRight = min(imwidth, offset + cX + window_width // 2)

        if self.annotation_json["annotations"]:
            new_ctr = []
            for i in range(len(seg)):
                if i % 2 == 0:
                    #new_ctr.append(seg[i] - newLeft - offset) # not sure why this is broken now
                    new_ctr.append(seg[i] - newLeft)
                else:
                    new_ctr.append(seg[i])
        else:
            new_ctr = None

        new_ctr = AnnotatedImage.coco_to_opencv(new_ctr)

        new_img = self.img[:, newLeft:newRight]
        mask = np.zeros(new_img.shape, np.uint8)
        cv2.drawContours(mask, new_ctr, -1, (255), 1)

        return AnnotatedImage(new_img, mask)


    def coco_to_opencv(segmentation):
        if type(segmentation) == type(None):
            return None

        points = [ [segmentation[i], segmentation[i+1]] for i in range(0, len(segmentation), 2) ]
        ctr = np.array(points).reshape((-1,1,2)).astype(np.int32)
        return ctr

    def opencv_to_coco(contour):
        if type(contour) == type(None):
            return None

        segmentation = []
        for contour in contours:
            contour = contour.flatten().tolist()
            contour_pairs = [(contour[i], contour[i+1]) for i in range(0, len(contour), 2)]
            segmentation.append([int(coord) for pair in contour_pairs for coord in pair])

        return segmentation



    # how to do image resizing?
    # just mutate the object


class ImageAnnotater:
    def __init__(self, image, resize_constant):
        self.image = image
        self.resize_constant = resize_constant

    def annotate(self):
        prepared_image = self._get_prepared_image(self.image)
        xmin, xmax, ymin, ymax = self._get_roi(prepared_image)
        mask = self._get_initial_mask(prepared_image, xmin, xmax, ymin, ymax)

        d = Drawing(prepared_image, mask)
        final_mask = d.draw()

        return AnnotatedImage(prepared_image, final_mask)

    def _get_initial_mask(self, prepared_image, xmin, xmax, ymin, ymax):
        lower, higher = 0, 255
        mask_view = False
        contour_view = True
        mask = None
        
        window_name = "Choose Initial Mask"
        cv2.namedWindow(window_name, cv2.WINDOW_KEEPRATIO)
        cv2.resizeWindow(window_name, (1920, 1080))
        cv2.imshow(window_name, prepared_image)

        while True:
            key = cv2.waitKey(1)
                    
            if key == 27:
                break
            elif key == ord('l'):
                lower += 1
            elif key == ord('k'):
                lower -= 1
            elif key == ord('h'):
                higher -= 1
            elif key == ord('j'):
                higher += 1
            elif key == ord('m'):
                mask_view = not mask_view
            elif key == ord('c'):
                contour_view = not contour_view
            elif key == ord('f'):
                print('finishing')
                break

            mask = self._get_mask(prepared_image, lower, higher, xmin, xmax, ymin, ymax)
            contours, hierarchy = cv2.findContours(mask, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
            frame = prepared_image.copy()

            if contour_view:
                cv2.drawContours(frame, contours, -1, (0, 255, 0), 1)

            frame = cv2.putText(frame, f"{lower}-{higher}", (0, 25), cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 0, 0), 2)
            if mask_view:
                cv2.imshow(window_name, mask)
            else:
                cv2.imshow(window_name, frame)

        cv2.destroyWindow(window_name)

        kernel = np.ones((3,3),np.uint8)
        mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE,kernel, iterations = 2)
        return mask

    def _get_roi(self, image):
        cv2.namedWindow("Select ROI", cv2.WINDOW_KEEPRATIO)
        cv2.resizeWindow("Select ROI", (1920, 1080))

        bbox = cv2.selectROI("Select ROI", image)
        cv2.destroyWindow("Select ROI")

        xmin = bbox[0]
        ymin = bbox[1]
        xmax = xmin + bbox[2]
        ymax = ymin + bbox[3]

        return xmin, xmax, ymin, ymax

    def _get_mask(self, image, lower, upper, xmin, xmax, ymin, ymax):
        thresh = cv2.inRange(image, lower, upper)
        bbox_mask = thresh.copy()
        bbox_mask[0:, 0:] = 0
        cv2.rectangle(bbox_mask, (xmin, ymin), (xmax, ymax), 255, -1)
        thresh_final = cv2.bitwise_and(thresh, bbox_mask)

        # noise removal
        kernel = np.ones((3,3),np.uint8)
        thresh_final = cv2.morphologyEx(thresh_final,cv2.MORPH_OPEN,kernel, iterations = 2)

        return thresh_final
        
    def _get_prepared_image(self, image):
        shape = image.shape
        image = cv2.resize(image, (shape[1] * self.resize_constant, shape[0] * self.resize_constant), cv2.INTER_NEAREST)
        imgray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        return imgray



from setuptools import find_packages, setup

setup(
    name="golgi-cell-cv",
    version="0.1",
    packages=find_packages(where="src"),
    package_dir= {"": "src"},
    package_data= {"": ["*/.gitkeep"]},
    description="description",
    url="https://github.com/AndKaminer/golgi",
    author="Andrew Kaminer",
    author_email="akaminer@gatech.edu",
    install_requires=[
        'opencv-python',
        'numpy',
        'roboflow',
        'easygui',
        'huggingface_hub',
        'ultralytics',
        'diplib'
    ]
)



from .data_handler import DataHandler
from .util import Util

import numpy as np
import cv2
import diplib as dip

class ProcessedImage:

    def __init__(self, frame, centerX, window_width, scaling_factor, um_per_pixel, model):
        
        self.frame = frame
        self.centerX = centerX
        self.window_width = window_width
        self.model = model
        self.scaling_factor = scaling_factor
        self.um_per_pixel = um_per_pixel

        self.data = {
                'area' : 0,
                'perimeter' : 0,
                'height' : 0,
                'circularity' : 0,
                'centerX' : None}
        
        self.contour = None
        self.contour = self.get_contour()
        self.process_contour(self.contour)


    def get_dip_measurement(self):
        binary_image = cv2.cvtColor(np.zeros_like(self.frame), cv2.COLOR_BGR2GRAY)

        cv2.drawContours(binary_image, [self.contour], -1, (1), thickness=cv2.FILLED)

        binary_image_dip = dip.Image(binary_image.astype(np.uint8))
        binary_image_dip.Convert('BIN')

        labeled_image = dip.Label(binary_image_dip)
        measurement = dip.MeasurementTool.Measure(labeled_image, features=['Perimeter', 'SolidArea', 'Roundness'])

        return measurement


    def get_contour(self):
        return self.contour


    def process_contour(self, contour):
        if contour is None:
            return

        M = cv2.moments(contour)
        m10, m00 = M["m10"], M["m00"]
        if m00 != 0:
            self.data['centerX'] = int(m10 // m00)
        
        _, _, _, h = cv2.boundingRect(contour)

        self.data['height'] = h * self.um_per_pixel / self.scaling_factor

        measurement = self.get_dip_measurement()

        self.data['perimeter'] = measurement['Perimeter'][1][0] * self.um_per_pixel / self.scaling_factor
        self.data['area'] = measurement['SolidArea'][1][0] * pow(self.um_per_pixel, 2) / pow(self.scaling_factor, 2)
        self.data['circularity'] = measurement['Roundness'][1][0] 


    def get_contour(self):
        if self.contour is not None:
            return self.contour

        width = self.frame.shape[1]

        new_window = Util.get_window(self.frame, self.centerX, self.window_width)

        if new_window is None:
            return None

        results = self.model(new_window, max_det=1, verbose=False)

        mask = []

        if not results[0].masks:
            return None

        for r in results:
            plot = r.plot(boxes=False)
            for x, y in r.masks.xy[0]:
                mask.append(int(x) + self.centerX - self.window_width)
                mask.append(int(y))
        ctr = np.array(mask).reshape((-1, 1, 2)).astype(np.int32)
        
        self.contour = ctr
        return ctr


    def get_parameter(self, parameter):
        return self.data[parameter]
